#ifndef __PASSCODE_NODE_HPP__
#define __PASSCODE_NODE_HPP__

#include "../params.hpp"
#include "../dataparser/structures.hpp"
#include "../dataparser/instance_loader.hpp"
#include "../utils/distributed_essentials.hpp"
#include "../utils/my_cblas_wrapper.hpp"
#include "../utils/utils.hpp"
#include "../lossfunction/loss_function.hpp"

template<typename L, typename D>
class BaselineFramework {
  public:
    BaselineFramework(Parameters &par, ProblemData<L,D> &pd, Solver<L,D> *s, LossFunction<L,D> *_lf, StatsLogger<L,D> &sl) : params(par), instance(pd), solver(s), lf(_lf), stats_logger(sl) {
      init();
    }
    virtual ~BaselineFramework() {
    }

  private:
    void init() {
      instance.lambda = params.lambda;
      wBuffer.resize(instance.m + 1);
      deltaW.resize(instance.m + 1); // piggyback local dual objective
      w.resize(instance.m+1);
      instance.x.resize(instance.n);
      cout << "x.size=" << instance.x.size() << endl;
      cblas_set_to_zero(instance.x);

      instance.total_n = instance.n;
      cout << "Local n " << instance.n << "   global n " << instance.total_n << endl;

      cblas_set_to_zero(w);
      cblas_set_to_zero(deltaW);

      instance.oneOverLambdaN = 1 / (0.0 + instance.total_n * instance.lambda);

      params.gamma = 1;
      instance.penalty = params.update_set + 0.0;

    }

public:

    void run() {
      
      double start = 0;
      double finish = 0;
      double elapsedTime = 0;
      unsigned long communication_count = 0;

      cout << "communicate count = " << params.bulk_iters << endl;
      cout << "bulk_iteration count = " << params.bulk_size << endl;
      cout << "per thread iterations count = " << params.local_iters << endl;

      for (unsigned int t = 0; t < params.bulk_iters; t++) {

        start = gettime_();

        for (int jj = 0; jj < params.bulk_size; jj++, communication_count++) {
          cblas_set_to_zero(deltaW);
          solver->solve(w, deltaW);
          cblas_sum_of_vectors(w, deltaW, params.gamma);
        }
        finish = gettime_();
        elapsedTime += finish - start;

        double dualError, primalError;
        lf->computeObjectiveValue(w, dualError, primalError);
        stats_logger.writeTempLog(instance, communication_count, elapsedTime, w, dualError, primalError);
      }

    }

  protected:
    Parameters& params;
    ProblemData<L,D> &instance;
    Solver<L,D> *solver;
    LossFunction<L,D> *lf;
    StatsLogger<L,D> &stats_logger;
    std::vector<D> wBuffer;
    std::vector<D> deltaW;
    std::vector<D> w;
};


#endif // __PASSCODE_NODE_HPP__

