#ifndef __MASTER_HPP__
#define __MASTER_HPP__

#include "node.hpp"

template<typename L, typename D>
class Master : public Node<L,D> {
  public:
    Master(boost::mpi::communicator& world, Parameters &par, ProblemData<L,D> &ins, Solver<L,D> *s, LossFunction<L,D> *lf, StatsLogger<L,D> &sl) : Node<L,D>(world, par, ins, s, lf, sl) {
    }
    ~Master() {}
    void run() {
      double start = 0;
      double finish = 0;
      double elapsedTime = 0;
      long grand_max_delay = 0;

      vector<D> deltaWOthers(instance.m+1); // loalDual is piggybacked
      vector<vector<D>> rcvBuffer(world.size());
      vector<D> dualObjs(world.size(), 0.0);
      vector<long> worker_rounds(world.size(), params.bulk_iters * params.bulk_size);
      vector<long> delay(world.size(), 0);
      int num_received = 0;

      for (int i=0; i<world.size(); ++i) rcvBuffer[i].resize(instance.m+1); // loalDual is piggybacked

      vector<boost::mpi::request> reqs;
      vector<boost::mpi::status> stats(world.size() - 1);
      for (int worker=1; worker < world.size(); worker++) {
        reqs.push_back(world.irecv(worker, 16 /*TAG*/, rcvBuffer[worker]));
      }

      unsigned long t=0;
      size_t workers_alive = world.size() - 1;
      size_t MINIMUM_UPDATES = min(workers_alive, params.update_set - 1);

      while (1) {

        start = gettime_();

        cblas_set_to_zero(deltaW);
        cblas_set_to_zero(deltaWOthers);
        solver->solve(w, deltaW);
        lf->computeDualObjective(dualObjs[0]); 

        int max_delay = 0;
        auto pos = reqs.end();
        do {
          //cout << world.rank() << " waiting for some " << endl;
          auto tmp_st = stats.begin()+num_received;
          tie(tmp_st, pos) = boost::mpi::wait_some(reqs.begin(), pos, tmp_st);
          for (; stats.begin()+num_received != tmp_st; ++num_received) {
            int worker = stats[num_received].source();
            //cout << "num_received = " << num_received << ":" << worker << endl;
            delay[worker] = 0;
            worker_rounds[worker]--;
          }
          max_delay = 0;
          for (int i=1; i<world.size(); ++i) {
            if (max_delay < delay[i]) 
              max_delay = delay[i];
          }
          if (max_delay > grand_max_delay) grand_max_delay = max_delay;
          //cout << "num_received = " << num_received << " vs min_updates=" << MINIMUM_UPDATES << endl;
          //cout << "max_delay = " << max_delay << " vs min_tau=" << params.update_tau << endl;
        } while ((num_received < MINIMUM_UPDATES) || (max_delay > params.update_tau));
        for (int i=0; i<MINIMUM_UPDATES; ++i) {
          int worker = stats[i].source();
          cblas_sum_of_vectors(deltaWOthers, rcvBuffer[worker]);
          dualObjs[worker] = rcvBuffer[worker][instance.m];
        }
        reqs.erase(pos, end(reqs));
        //cout << "stats.size = " << stats.size() << endl;

        //cout << "req.size=" << reqs.size() << endl;
        //for (int worker=1; worker < world.size(); ++worker) cout << worker_rounds[worker] << " ";
        //cout << endl;

        cblas_sum_of_vectors(deltaW, deltaWOthers);
        cblas_sum_of_vectors(w, deltaW, params.gamma);
        vector<boost::mpi::request> send_reqs;
        for (int i=0; i<MINIMUM_UPDATES; ++i) {
          int worker = stats[i].source();
          if (worker_rounds[worker] > 0) {
            //cout << world.rank() << " sending to " << worker << endl;
            send_reqs.push_back(world.isend(worker, 16 /*TAG*/, w));
            reqs.push_back(world.irecv(worker, 16 /*TAG*/, rcvBuffer[worker]));
          } else {
            workers_alive--;
            delay[worker] = -1;
          }
        }
        for (int i=1; i<world.size(); ++i) {
          if (delay[i] >= 0) delay[i] += 1;
        }
        int i=0; 
        for (int j=MINIMUM_UPDATES; j<num_received; ++j, ++i) {
          stats[i] = stats[j];
        }
        num_received = i;
        if (MINIMUM_UPDATES > workers_alive) MINIMUM_UPDATES = workers_alive;
        //cout << world.rank() << " waiting for all " << endl;
        boost::mpi::wait_all(begin(send_reqs), end(send_reqs));
        //vall_reduce(world, deltaW, wBuffer);
        //cblas_sum_of_vectors(w, wBuffer, gamma);
        //cblas_sum_of_vectors(w, deltaW, gamma);
        //cblas_sum_of_vectors(instance.x, deltaAlpha, gamma);
        if (t++ % params.bulk_size == params.bulk_size - 1) {
          double finalDualError;

          finish = gettime_();
          elapsedTime += finish - start;
          
          finalDualError = 0.0;
          for (int i=0; i<world.size(); ++i) finalDualError += dualObjs[i];
          //w[instance.m] = 0;
          stats_logger.writeTempLog(instance, t, elapsedTime, w, finalDualError);
        }
        if (!reqs.size()) break;
      }
    }
  private:
    using Node<L,D>::world;
    using Node<L,D>::params;
    using Node<L,D>::instance;
    using Node<L,D>::stats_logger;
    using Node<L,D>::wBuffer;
    using Node<L,D>::deltaW;
    using Node<L,D>::w;
    using Node<L,D>::solver;
    using Node<L,D>::lf;
};

#endif // __MASTER_HPP__
