#include <boost/mpi.hpp>
#include <getopt.h>
#include <string>

using namespace std;

#include "params.hpp"
#include "framework/master.hpp"
#include "framework/worker.hpp"
#include "framework/cocoa.hpp"
#include "framework/baseline.hpp"

typedef unsigned int L;
typedef double D;

void usage(const char *argv) {
  cout << "Usage: " << argv << " [OPTIONS] " << endl;
  cout << "\t-A <infile>  Prefix of the svm input file(s), actual filename must follow .<#nodes>.<node-id> " << endl;
  cout << "\t-B <int>     Number of times subproblem is solved in a bulk iteration" << endl;
  cout << "\t-f <int>     Loss function: " << LossFunctionID::HINGE << " for hinge" << endl;
  cout << "\t-H <int>     Number of local iterations by subproblem solver" << endl;
  cout << "\t-I <int>     Number of bulk iterations to be executed" << endl;
  cout << "\t-l <double>  Parameter lambda" << endl;
  cout << "\t-M <int>     Subproblem solver: " << SolverID::SDCA << " for SDCA, " << SolverID::PASSCODE << " for PassCoDe" << endl;
  cout << "\t-o <outfile> Output file : (default=\"" << default_out << "\")" << endl;
  cout << "\t-p <int>     Number of (parts of) input files : (default=16)" << endl;
  cout << "\t-S <int>     Size of update set : (default=num-nodes)" << endl;
  cout << "\t-T <int>     Delay bound (tau) : (default=\"" << 1 << "\")" << endl;
  cout << "\t-t <int>     Number of threads" << endl;
  cout << "\t-W <int>     Framework: " << FrameworkID::COCOA << " for CoCoA(+), " << FrameworkID::HDCA << " for Hybrid-DCA, " << FrameworkID::BASELINE << " for Baseline" << endl;
  exit(-1);
}

void parseCommandLine(int argc, char **argv, Parameters &params)
{
  int option_char;
  while ((option_char = getopt(argc, argv, "A:B:f:H:I:M:o:p:l:t:S:T:W:?")) != -1) {
    switch (option_char)
    {  
      case 'A': params.input_file_prefix = optarg; break;
      case 'B': params.bulk_size = atoll(optarg); break;
      case 'f': params.loss_function = static_cast<LossFunctionID>(atoi(optarg)); break;
      case 'H': params.local_iters = atoll(optarg); break;
      case 'I': params.bulk_iters = atoll(optarg); break;
      case 'l': params.lambda = atof (optarg); break;
      case 'M': params.sub_solver = static_cast<SolverID>(atoi(optarg)); break;
      case 'o': params.output_file_name = optarg; break;
      case 'p': params.input_num_parts = atoi (optarg); break;
      case 'S': params.update_set = atoi (optarg); break;
      case 'T': params.update_tau = atoi (optarg); break;
      case 't': params.num_threads = atoi (optarg); break;
      case 'W': params.framework = static_cast<FrameworkID>(atoi(optarg)); break;
      default: cerr << "unknown option " << (char)option_char << endl; usage(argv[0]); break;
    }
  }

  if (argc - optind < 0) {
    usage(argv[0]);
  } 
}


int main(int argc, char *argv[])
{
  boost::mpi::environment env{argc, argv};
  boost::mpi::communicator world;
  std::cout << world.rank() << ", " << world.size() << '\n';

  Parameters params;
  params.update_set = world.size();
  params.num_threads = omp_get_max_threads();

  //Statistics stats;  
  parseCommandLine(argc, argv, params);

  if (world.rank() == 0) {
  cout << "Parameters selected:" << endl;
  cout << "\tInput file         = '" << params.input_file_prefix << "'" << endl;
  cout << "\t# local iterations = " << params.local_iters << endl;
  cout << "\t# threads          = " << params.num_threads << endl;
  cout << "\tAPPROX             = " << params.APPROX << endl;
  cout << "\t(S, tau)           = " << params.update_set << "," << params.update_tau<< endl;
  cout << "\tloss function      = " << params.loss_function << endl;
  cout << "\tsubproblem solver  = " << params.sub_solver << endl;
  cout << "\tlambda             = " << params.lambda << endl;
  cout << "\tFramework          = " << params.framework << endl;
  cout << "\tOutput file        = " << params.output_file_name << endl;
  }

  omp_set_num_threads(params.num_threads);

  ProblemData<L,D> instance;
  loadDistributedSparseSVMRowData(params.input_file_prefix, world.rank(), world.size(), params.input_num_parts, instance, false);

  Solver<L,D> *solver = 0;
  LossFunction<L,D> *lf = 0;

  switch (params.sub_solver) {
    case SolverID::SDCA:
      solver = new Sdca<L,D>(instance, params);
      break;
    case SolverID::PASSCODE:
      solver = new PassCode<L,D>(instance, params);
      break;
    default:
      break;
  }

  switch (params.loss_function) {
    case LossFunctionID::HINGE:
      lf = new HingeLoss<L,D>(instance, params);
      break;
    default:
      break;
  }
  bool use_temp_log = (params.framework == FrameworkID::HDCA);
  StatsLogger<L,D> sl(world, params, lf, use_temp_log);

  if (world.rank() == 0) {
    params.out.open(params.output_file_name.c_str());
    if (!params.out) {
      cout << "Could not open file : " << params.output_file_name << endl;
      exit(-1);
    }
  }

  switch(params.framework) {
    case FrameworkID::COCOA:
      {
        CoCoA<L,D> cocoa(world, params, instance, solver, lf, sl);
        lf->resetInstance();
        cocoa.run();
        break;
      }
    case FrameworkID::HDCA:
      {
        if (world.rank() == 0) {
          Master<L,D> master(world, params, instance, solver, lf, sl);
          lf->resetInstance();
          master.run();
        } else {
          Worker<L,D> worker(world, params, instance, solver, lf, sl);
          lf->resetInstance();
          worker.run();
        }
        break;
      }
    case FrameworkID::BASELINE:
      {
        if (world.rank() == 0) {
          cout << "Baseline can be run only on a single node! Running on root only." << endl;
          BaselineFramework<unsigned int, double> pf(params, instance, solver, lf, sl);
          lf->resetInstance();
          pf.run();
        }
        break;
      }
  }

  world.barrier();

  if (use_temp_log) {
    cout << "[" << world.rank() << "] finalizing log ....." << endl;
    //loadDistributedSparseSVMRowData(params.input_file_prefix, 0, 1, params.input_num_parts, instance, false);
    sl.writeFinalLog(instance);
  }

  if (world.rank() == 0) {
    params.out.close();
  }
  delete lf;
  delete solver;
}
