#ifndef __STATS_LOGGER_H__
#define __STATS_LOGGER_H__

#include "../params.hpp"
#include "../dataparser/structures.hpp"
#include "../lossfunction/loss_function.hpp"

#define TEMP_FILE "temp_stat_log.bin"

template<typename L, typename D>
class StatsLogger {
public:
  StatsLogger(boost::mpi::communicator& _world, Parameters &par, LossFunction<L,D> *_lf, bool utd=false) : world(_world), params(par), file(0), lf(_lf), num_records(0), use_temp_disk(utd), wsize(0) {
    //cout << "Entered stats logger constructor" << ", rank=" << world.rank() << endl;
    if (use_temp_disk && (world.rank() == 0)) {
      //cout << "in master before open" << endl;
      file = fopen(TEMP_FILE, "wb+");
      //cout << "in master after open" << endl;
      if (!file)
      {
        cout << "Error: Cannot open temporary file " << TEMP_FILE << "\n";
        exit(1);
      }
      //cout << "Left stats logger constructor(MASTER)" << endl;
    } else {
      //cout << "Left stats logger constructor(WORKER)" << endl;
    }
  }
  ~StatsLogger() {
  }

  void writeTempLog(ProblemData<L,D> &instance, L iter, D elapsedTime, vector<D> &w, D dualError, D primalError=0.0) {
    cout << "Iteration " << iter << " elapsed time " << elapsedTime << "  error " << primalError<< " dual " << dualError << endl;
    if (use_temp_disk) {
      wsize = w.size();
      fwrite(&iter, sizeof(L), 1, file);
      fwrite(&elapsedTime, sizeof(D), 1, file);
      fwrite(&w[0], sizeof(D), wsize, file);
      fwrite(&dualError, sizeof(D), 1, file);
      num_records++;
    } else {
      writeOutput(instance, iter, elapsedTime, w, dualError, primalError);
    }
  }
  void writeFinalLog(ProblemData<L,D> &instance) {
    L iter;
    D elapsedTime;
    D dualError;
    if (file) {
      rewind(file);
    } else {
      file = fopen(TEMP_FILE, "rb");
    }
    size_t final_num_records = 0, final_wsize = 0;
    vall_reduce(world, &num_records, &final_num_records, 1);
    vall_reduce(world, &wsize, &final_wsize, 1);
    vector<D> w(final_wsize);
    cout << "In Write final log ... " << world.rank() << ", wsize=" << final_wsize << endl;
    for (size_t i=0; i<final_num_records; ++i) {
      fread(&iter, sizeof(L), 1, file);
      fread(&elapsedTime, sizeof(D), 1, file);
      fread(&w[0], sizeof(D), final_wsize, file);
      fread(&dualError, sizeof(D), 1, file);
      D primalError, finalPrimalError = 0.0;
      lf->computePrimalObjective(instance, w, primalError);
      //cout  << "[" << world.rank() << "]" << "primal error=" <<primalError << ", finalPrimalError=" << finalPrimalError << endl;
      //cout << "[" << world.rank() << "]" << w[0] << "\t" << w[1] << "\t" << w[117] << "\t" << w[118] << endl;
      vall_reduce(world, &primalError, &finalPrimalError, 1);
      cout  << "[" << world.rank() << "]" << "primal error=" <<primalError << ", finalPrimalError=" << finalPrimalError << endl;
      if (world.rank() == 0) {
        writeOutput(instance, iter, elapsedTime, w, dualError, finalPrimalError);
      }
    }
  }
private:
  void writeOutput(ProblemData<L,D> &instance, L iter, D elapsedTime, vector<D> &w, D dualError, D primalError) {
    cout << "dual=" << dualError << ", primal=" << primalError << endl;

    D tmp2 = cblas_l2_norm(w.size()-1, &w[0], 1);
    cout << "tmp2=" << tmp2 << endl;
    dualError = 1 / (0.0 + instance.total_n) * dualError
      + 0.5 * instance.lambda * tmp2 * tmp2;
    primalError =  1 / (0.0 + instance.total_n) * primalError
      + 0.5 * instance.lambda * tmp2 * tmp2;

    cout << "Iteration " << iter << " elapsed time " << elapsedTime
      << "  error " << primalError<< "    " << dualError
      << "    " << primalError + dualError << endl;

    params.out << iter << "," << elapsedTime << "," << primalError << ","
      << dualError << "," << primalError + dualError << endl;
  }
private:
  boost::mpi::communicator& world;
  Parameters &params;
  FILE* file;
  LossFunction<L,D> *lf;
  size_t num_records;
  bool use_temp_disk;
  size_t wsize;
};

#endif // __STATS_LOGGER_H__

